// copyright 2002 Josh Levenberg, please see copying.txt
#include "fractal.h"
#include "hf.h"
#include "writepng.h"
#include "vect3.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
enum { HeightFieldShift=12 };
enum { HeightFieldSize=(1<<HeightFieldShift) };
*/
static int sShift;
static int sSize;
static float *sHeightField;
static char *sHeightName;
static char *sTexName;

void GenerateHeightField(float *heightField, int randomize)
  {
    assert((1<<sShift)==sSize);
    WrappingHeightField hf(sShift);
    if (randomize)
      WrappingHeightField::SeedRandomizer(randomize);
    hf.JFractal(1<<24, 1.414, 1<<24, 1<<30);
    int minH, maxH;
    hf.GetMinMax(minH, maxH);
    float factor=HEIGHT_FIELD_MAX_HEIGHT/(maxH-minH+1);
    for (WrappingHeightField::iterator cur=hf.Begin(); cur!=hf.End(); ++cur)
      *heightField++ = (*cur-minH)*factor;
  }

void WriteHeightField()
  {
    printf("Writing height field to disk as %s\n", sHeightName);
    unsigned short *image=new unsigned short[(sSize+1)*(sSize+1)];
    for (int y=0; y<=sSize; ++y)
      for (int x=0; x<=sSize; ++x)
        {
          image[y*(sSize+1)+x]=sHeightField[(y&(sSize-1))*sSize+(x&(sSize-1))]
                                *65535/HEIGHT_FIELD_MAX_HEIGHT;
        }
    bool result=write_png_16gray(sHeightName, sSize+1, sSize+1, image);
    printf("Result = %s\n", result ? "true" : "false");
  }

void GenerateTextureMap()
  {
    unsigned char *gray=new unsigned char[sSize*sSize];

#define HEIGHT(x,y) sHeightField[((y)&(sSize-1))*sSize+((x)&(sSize-1))]

    vec3_t vn;
    float dx=1.f/sSize;
    for (int y=0; y<sSize; ++y)
      for (int x=0; x<sSize; ++x)
        {
          float v1 = HEIGHT(x+1,y+1)-HEIGHT(x,y);
          float v2 = HEIGHT(x+1,y)-HEIGHT(x,y+1);

          // cross product to find normal
          vn[0] = dx*v2+v1*dx;
          vn[1] = v1*dx-dx*v2;
          vn[2] = -2*dx*dx;

          // make normal have unit length
          vnormalize(vn);

          // dot product with negative of light vector (1,1,1)/sqrt(3) to find color
          float clr = (-vn[0]-vn[1]-vn[2])/1.7321;
          if (clr<0)
            clr=0;

          // Use 20% ambient, convert to 0-255
          gray[y*sSize+x]=(.8*clr+.2)*255;
        }

#undef HEIGHT

    printf("Writing texture map to disk as %s\n", sTexName);
    bool result=write_png_8gray(sTexName, sSize, sSize, gray);
    printf("Result = %s\n", result ? "true" : "false");

    // clean up
    delete[] gray;
  }

int main(int argc, char *argv[])
  { // first parameter = hf shift
    // second parameter = filename
    // optional third parameter = seed

    if (argc<2)
      {
        printf("Usage: %s shift [filename [seed]]\n"
               "    will make a 2^shift+1 square height field\n"
               "    will write it to filename_hf.png\n"
               "    will write a texture map to filename_tex.png\n"
               "    filename defaults to \"def\"\n"
               "    every seed gives a different height field (defaults to 42)\n",
               argv[0]);
        return 2;
      }
    
    sShift=atoi(argv[1]);
    if (sShift<2 || sShift>16)
      {
        printf("Invalid shift value of %d", sShift);
        return 1;
      }
    sSize=1<<sShift;

    const char *base_name="def";
    if (argc>2)
      base_name=argv[1];
    int len=strlen(base_name);

    sHeightName=new char[len+8];
    strcpy(sHeightName, base_name);
    strcat(sHeightName, "_hf.png");
    
    sTexName=new char[len+9];
    strcpy(sTexName, base_name);
    strcat(sTexName, "_tex.png");

    int seed=42;
    if (argc>3)
      seed=atoi(argv[3]);

    printf("Generating height field.  Please wait.\n");

    sHeightField = new float[sSize*sSize];
    GenerateHeightField(sHeightField, seed);

    WriteHeightField();

    printf("Generating texture map.\n");
    GenerateTextureMap();

    delete[] sHeightField;
    sHeightField=0;

    delete[] sHeightName;
    delete[] sTexName;

    return 0;
  }

