// copyright 2002 Josh Levenberg, please see copying.txt
/* see model.h for some comments */
#include "model.h"
#include "cachehlp.h"
#include "dlod.h"
#include "hf.h"

using namespace CacheHelper;

//Allow cracks instead of uniformly subdividing boundaries:
//#define CRACKS  // UNTESTED!

// global data ----------------------------------------------------------------
Model *gModel=0;
Errors *gErrors=0;
AggrErrors *gAggrErrors=0;

// local class declarations ----------------------------------------------------------------

class SubdivTri;

class AggrErrorsTAdaptive : public AggrErrorsT
  {
protected:
    friend class SubdivTri;
    int mAllocatePoolSize;

    std::auto_array<SubdivTri> mSubdivTriPool;
    SubdivTri *mSubdivTriPoolEnd;

    std::auto_array<SubdivTri *> mRadiusQueueBegin;
    SubdivTri **mRadiusQueueEnd;

    vec_t AdaptiveTreeHelper(ind lind, ind rind, ind bind, int &numtri);
    vec_t RadiusForAdaptiveTree(ind, ind, ind);
    int ConstructAdaptiveTree(ind, ind, ind);
    void RenderAdaptiveTree();
    virtual vec_t RadiusForTree(ind a, ind b, ind c)
      { return RadiusForAdaptiveTree(a,b,c); }
    virtual int ConstructTree(ind a, ind b, ind c)
      { return ConstructAdaptiveTree(a,b,c); }
    virtual void RenderTree()
      { RenderAdaptiveTree(); }
    friend class ErrorsT;
    AggrErrorsTAdaptive(int aggr);

public:
    virtual bool GetAdaptive()
      { return true; }

    // for use through sAdaptive
    SubdivTri *AllocateTwoSubdivTris();
    void InitRadiusPriorityQueue(SubdivTri *tri);
    void EnqueueRadiusPriorityQueue(SubdivTri *tri);
    SubdivTri *PopRadiusPriorityQueue();
    vec_t PeekTopRadiusPriorityQueue();
    inline int SplitsLeftInPool();
    inline int NumLeavesInPool();
  };

static AggrErrorsTAdaptive *sAdaptive=0;

class AggrErrorsTUniform : public AggrErrorsT
  {
protected:
    virtual vec_t RadiusForTree(ind a, ind b, ind c)
      { return RadiusForUniformTree(a,b,c); }
    virtual int ConstructTree(ind a, ind b, ind c)
      { return ConstructUniformTree(a,b,c); }
    virtual void RenderTree()
      { RenderUniformTree(); }
    friend class ErrorsT;
    AggrErrorsTUniform(int aggr);

public:
    virtual bool GetAdaptive()
      { return false; }
  };

// inline helper functions ----------------------------------------------------------------

static inline void maxify(vec_t &a, vec_t &b)
  {
    if (a>b)
      b=a;
    else
      a=b;
  }

static inline float lerp(float r, float l, float h)
  { 
    return l + (h-l)*r; 
  }

static inline vec_t Distance(vec3_t a, vec3_t b)
  {
    return sqrt(vdistsq(a, b));
  }

// global functions ----------------------------------------------------------------

void InitializeModel(std::auto_array<unsigned short> hf, vec_t maxH, int shift)
  {
    delete gModel;
    gModel = new ModelT(hf, maxH, shift);
  }

void InitializeErrors(vec_t error_ratio)
  {
    assert(gModel);
    delete gErrors;
    gErrors=0;
    if (gAggrErrors)
      {
        delete gAggrErrors;
        gAggrErrors=0;
      }
    gErrors=gModel->CalculateErrors(error_ratio).release();
  }

void InitializeAggregation(int aggr, bool adapt)
  {
    assert(gModel);
    assert(gErrors);
    if (gAggrErrors)
      if (gAggrErrors->GetAggregation()==aggr && gAggrErrors->GetAdaptive()==adapt)
        return;
      else
        {
          delete gAggrErrors;
          gAggrErrors=0;
        }
    gAggrErrors=gErrors->ComputeAggregation(aggr, adapt).release();
  }

// ModelT ----------------------------------------------------------------

vec_t ModelT::LookUpHeight(vec_t x, vec_t y)
  { // 0 <= x,y <= 1
    int xl=x*mSize, yt=y*mSize;
    float rx=x*mSize-xl,
          ry=y*mSize-yt;
    xl %= mSize+1;
    yt %= mSize+1;
    if (xl<0)
      xl+=mSize+1;
    if (yt<0)
      yt+=mSize+1;

#define HEIGHT(xx, yy) (mH[(yy)*(mSize+1)+(xx)]*mHScale)
    float hlt = HEIGHT(xl, yt),
          hrt = HEIGHT(xl+1, yt),
          hlb = HEIGHT(xl, yt+1),
          hrb = HEIGHT(xl+1, yt+1);
#undef HEIGHT
    return lerp(ry, lerp(rx, hlt, hrt), lerp(rx, hlb, hrb));
  }

std::auto_ptr<Errors> ModelT::CalculateErrors(vec_t error_ratio) const
  { 
    assert(gModel==this);
    return new ErrorsT(error_ratio); 
  }

void ModelT::BoundingVolume(ind left, ind right, ind base, 
                                    std::vector<vec3s_t> &ret)
  {
    ret.reserve(ret.size()+6);

#define mMinH (0.0)
    vec_t x=Xof(left)*mXYScale;
    vec_t y=Yof(left)*mXYScale;
    ret.push_back();
    ret.back().m[0]=x;
    ret.back().m[1]=y;
    ret.back().m[2]=mMinH;
    ret.push_back();
    ret.back().m[0]=x;
    ret.back().m[1]=y;
    ret.back().m[2]=mMaxH;
    x=Xof(right)*mXYScale;
    y=Yof(right)*mXYScale;
    ret.push_back();
    ret.back().m[0]=x;
    ret.back().m[1]=y;
    ret.back().m[2]=mMinH;
    ret.push_back();
    ret.back().m[0]=x;
    ret.back().m[1]=y;
    ret.back().m[2]=mMaxH;
    x=Xof(base)*mXYScale;
    y=Yof(base)*mXYScale;
    ret.push_back();
    ret.back().m[0]=x;
    ret.back().m[1]=y;
    ret.back().m[2]=mMinH;
    ret.push_back();
    ret.back().m[0]=x;
    ret.back().m[1]=y;
    ret.back().m[2]=mMaxH;
#undef mMinH
  }


// ErrorsT ----------------------------------------------------------------

vec_t ErrorsT::ErrorThresholdHelper(vec_t xn, vec_t yn, vec_t offset, vec3_t refpt,
                           int xb, int yb, int xl, int yl, int xr, int yr)
  {
    int xc=(xl+xr)>>1,
        yc=(yl+yr)>>1;
    vec3_t cpos;
    gModel->GetPosition(xc, yc, cpos);
    vec_t errdist = fabs(cpos[2]-(xn*cpos[0]+yn*cpos[1]+offset))*mErrorRatio;
    errdist += sqrt(vdistsq(cpos, refpt));

    if (abs(xb-xl)>1 || abs(yb-yl)>1)
      { // recurse
        vec_t cmpto=ErrorThresholdHelper(xn, yn, offset, refpt, xc, yc, xb, yb, xl, yl);
        errdist=max(errdist, cmpto);
        cmpto=ErrorThresholdHelper(xn, yn, offset, refpt, xc, yc, xr, yr, xb, yb);
        errdist=max(errdist, cmpto);
      }
    return errdist;
  }

void ErrorsT::ComputeErrorThreshold(int xb, int yb, int xl, int yl, int xr, int yr)
  {
    vec3_t planenormal;
    vec3_t bpos, lpos, rpos;
    gModel->GetPosition(xb, yb, bpos);
    gModel->GetPosition(xl, yl, lpos);
    gModel->GetPosition(xr, yr, rpos);

    { // planenormal = (gModel->mVerts[yb][xb] - gModel->mVerts[yl][xl]) x (gModel->mVerts[yr][xr] - gModel->mVerts[yb][xb])
      vec3_t t1, t2;
      vdiff(t1, bpos, lpos);
      vdiff(t2, rpos, bpos);
      vcross(planenormal, t1, t2);
    }

    vec_t xn = -planenormal[0] / planenormal[2];
    vec_t yn = -planenormal[1] / planenormal[2];

    vec_t offset = bpos[2]-(xn*bpos[0]+yn*bpos[1]);

    int xc=(xl+xr)>>1,
        yc=(yl+yr)>>1;
    vec3_t cpos;
    gModel->GetPosition(xc, yc, cpos);
    vec_t newerrdist=ErrorThresholdHelper(xn, yn, offset, cpos,
                                          xb,yb,xl,yl,xr,yr);
    mErrorThreshold[yc*(mSize+1)+xc] = max(mErrorThreshold[yc*(mSize+1)+xc], newerrdist);

    if (abs(xb-xl)>1 || abs(yb-yl)>1)
      { // recurse
        ComputeErrorThreshold(xc, yc, xb, yb, xl, yl);
        ComputeErrorThreshold(xc, yc, xr, yr, xb, yb);
      }
  }

// probably does 2x too much work compared to a bottom up algorithm
// (from visiting vertices twice as the hypotenuse of two different triangles)
// (it deals with squares as separate triangles)
// plus a bunch of recursion overhead
void ErrorsT::InitErrorThreshold()
  {
    // initialize to zero
    memset(mErrorThreshold.get(), 0, sizeof(vec_t)*(mSize+1)*(mSize+1));
    // (from now on will do mErrorThreshold[foo] = max(mErrorThreshold[foo], bar); )

    // compute naive error distances
    ComputeErrorThreshold(mSize, 0,
                          0,     0,
                          mSize, mSize);
    ComputeErrorThreshold(0,     mSize,
                          mSize, mSize,
                          0,     0);
  }

std::auto_ptr<AggrErrors> ErrorsT::ComputeAggregation(int agglevel, bool adaptive) const
  {
    assert(gErrors==this);
    if (adaptive)
      return new AggrErrorsTAdaptive(agglevel);
    else
      return new AggrErrorsTUniform(agglevel);
  }

// AggrErrorsT ----------------------------------------------------------------

AggrErrorsT::AggrErrorsT(int aggr) : mAggregation(aggr)
  {
    assert(mAggregation>=0);
    assert(mAggregation<=gModel->mShift);
    mAggrNumTri = 1 << (2*mAggregation);
    // allocate and compute error radii
    // fewer at high subdivision levels!
    mLowDetailTessellationShift=(gModel->mShift-mAggregation);
    mLowDetailTessellation=(1<<mLowDetailTessellationShift);
    mYShift=mAggregation+gModel->mShift+1;
  }

// low detail error computation ----------------

void AggrErrorsT::ComputeLowDetailErrorThreshold(int xb, int yb, int xl, int yl, int xr, int yr)
  {
    int xc=(xl+xr)>>1,
        yc=(yl+yr)>>1;

    ind bind = GetHD(GetVertexIndex(xb, yb));
    ind lind = GetHD(GetVertexIndex(xl, yl));
    ind rind = GetHD(GetVertexIndex(xr, yr));

    ind cind = GetHD(GetVertexIndex(xc, yc));
    vec_t olderrdist=gErrors->Get(cind);

    vec_t newerrdist=RadiusForTree(lind, rind, bind);

    vec_t *toupd=& Ref(yc, xc);
    *toupd = max(*toupd, newerrdist);

    if (abs(xb-xl)>(1<<mAggregation) || abs(yb-yl)>(1<<mAggregation))
      { // recurse
        ComputeLowDetailErrorThreshold(xc, yc, xb, yb, xl, yl);
        ComputeLowDetailErrorThreshold(xc, yc, xr, yr, xb, yb);
      }
  }

void AggrErrorsT::InitLowDetailErrorThreshold()
  {
    fprintf(gOutFile, "Aggregation level %d (%s):\n", mAggregation, 
            (GetAdaptive() ? "adaptive" : "uniform"));
    printf("Aggregation level %d (%s):\n", mAggregation, 
            (GetAdaptive() ? "adaptive" : "uniform"));

    // allocate error threshold
    int numerrors=2*(mLowDetailTessellation+1)*(mLowDetailTessellation+1);
    mLowDetailErrorThreshold.reset(new vec_t[numerrors]);

    // compute error radii

    // initialize to zero
    memset(mLowDetailErrorThreshold.get(), 0, sizeof(vec_t)*numerrors);
    // (from now on will do gLowDetailErrorThreshold[foo] = max(gLowDetailErrorThreshold[foo], bar); )

    // compute naive error distances
    ComputeLowDetailErrorThreshold(
                          gModel->mSize,  0,
                          0,              0,
                          gModel->mSize,  gModel->mSize);
    ComputeLowDetailErrorThreshold(
                          0,              gModel->mSize,
                          gModel->mSize,  gModel->mSize,
                          0,              0);
  }

// uniform subdivision ----------------

AggrErrorsTUniform::AggrErrorsTUniform(int aggr) : AggrErrorsT(aggr)
  {
    InitLowDetailErrorThreshold();
  }

int AggrErrorsT::ConstructUniformTree(ind lind, ind rind, ind bind)
  {
    vec lvec = Model::Diff(bind, lind);
    vec bvec = Model::Diff(rind, lind);
    int i;
    for (i=0; i<mAggregation; ++i)
      {
        lvec = Model::Halve(lvec);
        bvec = Model::Halve(bvec);
      }
    ind outind = lind;
    int count = 1 << mAggregation;
    for (i=0; i<=count; ++i)
      {
        ind inind = outind;
        for (int j=0; j<=count-i; ++j)
          {
            // check inind
            vec3_t vertpos;
            gModel->GetPosition(inind, vertpos);
            AllocateVertex(vertpos);

            // update inind
#ifndef NDEBUG
            if (j<count-i)  // trips an assert
#endif
              inind = Model::Add(inind, lvec);
          }
#ifndef NDEBUG
        if (i<count)  // trips an assert
#endif
          outind = Model::Add(outind, bvec);
      }

    return 1 << (2*mAggregation);
  }

vec_t AggrErrorsT::RadiusForUniformTree(ind lind, ind rind, ind bind)
  { 
    vec_t maxr=0;
    ind cind = Model::BisectUnsafe(lind, rind);
    vec3_t center;
    gModel->GetPosition(cind, center);
    vec lvec = Model::Diff(bind, lind);
    vec bvec = Model::Diff(rind, lind);
    int i;
    for (i=0; i<mAggregation; ++i)
      {
        lvec = Model::Halve(lvec);
        bvec = Model::Halve(bvec);
      }
    ind outind = Model::Add(lind, Model::Halve(bvec));
    int count = 1 << mAggregation;
    for (i=0; i<count; ++i)
      {
        ind inind = outind;
        for (int j=0; j<count-i; ++j)
          {
            // check inind
            vec3_t vertpos;
            gModel->GetPosition(inind, vertpos);
            vec_t rad = gErrors->Get(inind);
            rad += Distance(vertpos, center);
            maxr = max(maxr, rad);

            // update inind
#ifndef NDEBUG
            if (j+1<count-i)  // trips an assert
#endif
              inind = Model::Add(inind, lvec);
          }
#ifndef NDEBUG
        if (i+1<count)  // trips an assert
#endif
          outind = Model::Add(outind, bvec);
      }
    return maxr;
  }

void AggrErrorsT::RenderUniformTree()
  {
    if (! mAggregation)
      {
        PutTri(0, 2, 1);
        return;
      }

    int count = 1 << (mAggregation-1);
    int ind1 = 0;
    int ind2 = count + count + 1;
    int ind3 = ind2 + ind2 - 1;

    do 
      {
        int i=1;
        // do a half fan (4 triangles)

        // ind2 = center
        PutTri(ind2, ind1+1, ind1);
        PutTri(ind2, ind1+2, ind1+1);
        PutTri(ind2, ind2+1, ind1+2);
        PutTri(ind2, ind3, ind2+1);

        for (; i<count; ++i)
          {
            ind1 += 2;
            ind2 += 2;
            ind3 += 2;

            // do a whole fan (8 triangles)
            // ind2 = center
            PutTri(ind2, ind1, ind2-1);
            PutTri(ind2, ind2-1, ind3-2);
            PutTri(ind2, ind3-2, ind3-1);
            PutTri(ind2, ind3-1, ind3);
            PutTri(ind2, ind3, ind2+1);
            PutTri(ind2, ind2+1, ind1+2);
            PutTri(ind2, ind1+2, ind1+1);
            PutTri(ind2, ind1+1, ind1);
          }
        
        ind1 = ind2 + 2;
        ind2 = ind1 + (ind3-ind2) - 1;
        ind3 = ind2 + (ind2-ind1) - 1;
      } while (--count);
  }

#ifndef NDEBUG
bool AggrErrorsT::GoodIndex(ind i)
  {
    if (! gModel->GoodIndex(i))
      return false;
    int x=gModel->Xof(i);
    int y=gModel->Yof(i);
    return (x & ((1<<mAggregation)-1))==0 && 
           (y & ((1<<mAggregation)-1))==0;
  }
#endif

// SubdivTri -----------------------------------------------------------

struct SubdivLeaf
  {
    int mVertices[3];        // left, right, and base vertices
    int mCachedVert[3];
  };

struct SubdivBranch
  {
    int      mNegative;   // negative if a branch, >=0 if a leaf
    SubdivTri *mChildren;   // mChildren[0] is left, mChildren[1] is right
  };

class SubdivTri
  { // 32 bytes
protected:
    union
      {
        SubdivLeaf   uLeaf;
        SubdivBranch uBranch;  // Invariant: IsLeaf() || IsDummyBranch() ||
                             //      (uBranch.mChildren[0].mParent==this &&
                             //       uBranch.mChildren[1].mParent==this)
      };
    SubdivTri *mNeighbors[3];  
    // invariant: if this is a leaf, mNeighbors always points to other leaves
    // only mNeighbors[2] is accurate in Branch nodes
    int      mHypotVert;  // vertex in the midpoint of base edge (hypotenuse)
                          // can use to get radius of isosphere, -1 if undefined
    vec_t    mEffectiveRadius;  // used by radius priority queue

public:
    enum { Left, Right, Base };

    // Leaf functions
    bool IsLeaf() const { return uBranch.mNegative>=0; }
    void SetLeaf(int vertices[3])  
                  { assert(vertices[0]>=0); 
                    assert(vertices[1]>=0);
                    uLeaf.mVertices[0]=vertices[0];
                    uLeaf.mVertices[1]=vertices[1];
                    uLeaf.mVertices[2]=vertices[2];
                  }
    int & Vertex(int i) 
                  { assert(IsLeaf()); 
                    assert(i>=0);
                    assert(i<3);
                    return uLeaf.mVertices[i]; }
    int & CachedVertex(int i) 
                  { assert(IsLeaf()); 
                    assert(i>=0);
                    assert(i<3);
                    return uLeaf.mCachedVert[i]; }

    // Branch Functions
    void SetBranch(SubdivTri *children)
                  { assert(children);
                    uBranch.mNegative=-1;
                    uBranch.mChildren=children; }
    SubdivTri & Child(int i)
                  { assert(!IsLeaf());
                    assert(i>=0);
                    assert(i<2);
                    assert(uBranch.mChildren);
                    return uBranch.mChildren[i]; }
    SubdivTri * GetChildren()
                  { assert(!IsLeaf());
                    assert(uBranch.mChildren);
                    return uBranch.mChildren; }

    // Common accessor functions
    SubdivTri * Neighbor(int i) // if branch node, only use Base neighbor
                  { assert(i>=0);
                    assert(i<3);
                    assert(mNeighbors[i]==0 || IsLeaf() == mNeighbors[i]->IsLeaf());
                    assert(i==Base || IsLeaf());
                    return mNeighbors[i]; }
    void SetNeighbor(int i, SubdivTri *tri) // if branch node, only use Base neighbor
                  { assert(i>=0);
                    assert(i<3);
                    assert(tri==0 || IsLeaf() == tri->IsLeaf());
                    assert(i==Base || IsLeaf());
                    mNeighbors[i]=tri; }
    int & HypotVertIndex() 
                  { return mHypotVert; }
    vec_t &EffectiveRadius()  // used by radius priority queue
                  { return mEffectiveRadius; }

    //-- rendering
    void Render0();
    void Render1();

    // Split
    bool Split(int SplitCountDown=-1);

    void SplitEdge(int side);

protected:
    void SplitHelper(int CachedHypot);
  };


// Render
void SubdivTri :: Render0()
  {
    if (IsLeaf())
      {
        PutTri(CachedVertex(Left),  CachedVertex(Right), CachedVertex(Base));
      }
    else
      {
        Child(Left). Render1();
        Child(Right).Render1();
      }
  }

void SubdivTri :: Render1()
  {
    if (IsLeaf())
      {
        PutTri(CachedVertex(Left), CachedVertex(Right), CachedVertex(Base));
      }
    else
      {
        Child(Right).Render0();
        Child(Left). Render0();
      }
  }

// Split
bool SubdivTri :: Split(int SplitCountDown)
  {
    SubdivTri *base=Neighbor(Base);

    if (base && base->Neighbor(Base)!=this)
      { // need to force split base neighbor
        if (SplitCountDown==1 || ! Neighbor(Base)->Split(SplitCountDown-1))
          return false;
        assert(Neighbor(Base)->Neighbor(Base)==this);
        base=Neighbor(Base);
      }
#ifndef CRACKS
    else if (!base && SplitCountDown>0)
      return false;
#endif

    int CachedHypotVert;
    if (gIsCaching)
      {
        vec3_t CachedHypotVertPos;
        gModel->GetPosition(HypotVertIndex(), CachedHypotVertPos);
        CachedHypotVert=AllocateVertexNum(CachedHypotVertPos);
      }
    else
      CachedHypotVert=-1;

    SplitHelper(CachedHypotVert);

    if (base)
      {
        base->SplitHelper(CachedHypotVert);
        Child(Left).SetNeighbor(Right, & base->Child(Right));
        base->Child(Right).SetNeighbor(Left, & Child(Left));
        Child(Right).SetNeighbor(Left, & base->Child(Left));
        base->Child(Left).SetNeighbor(Right, & Child(Right));
      }
    else
      {
        Child(Left).SetNeighbor(Right, 0);
        Child(Right).SetNeighbor(Left, 0);
      }

    return true;
  }

inline SubdivTri * AggrErrorsTAdaptive::AllocateTwoSubdivTris()
  {
    SubdivTri *ret=mSubdivTriPoolEnd;
    mSubdivTriPoolEnd+=2;
    assert(mSubdivTriPoolEnd-mSubdivTriPool.get() <= mAllocatePoolSize);
    return ret;
  }

static vec3_t sCentralVertex;

static bool CompareSubdivTri(SubdivTri *l, SubdivTri *r)
  {
    return l->EffectiveRadius() < r->EffectiveRadius();
  }

// uses a raw array, make_heap, push_heap, pop_heap

void AggrErrorsTAdaptive::InitRadiusPriorityQueue(SubdivTri *tri)
  {
    //gRadiusQueue.clear();
    mRadiusQueueEnd=mRadiusQueueBegin.get();
    ind center=tri->HypotVertIndex();
    if (center>=0)
      {
        vec_t rad=gErrors->Get(center);
        gModel->GetPosition(center, sCentralVertex);
        tri->EffectiveRadius() = rad;
        //gRadiusQueue.push(tri);
        mRadiusQueueEnd=mRadiusQueueBegin+1;
        *mRadiusQueueBegin = tri;
      }
  }

void AggrErrorsTAdaptive::EnqueueRadiusPriorityQueue(SubdivTri *tri)
  {
    ind center=tri->HypotVertIndex();
    if (center>=0)
      {
        assert(mRadiusQueueEnd-mRadiusQueueBegin.get() < (mAggrNumTri<<1));
        vec_t rad=gErrors->Get(center);
        vec3_t vertpos;
        gModel->GetPosition(center, vertpos);
        rad += Distance(vertpos, sCentralVertex);
        tri->EffectiveRadius() = rad;
        //gRadiusQueue.push(tri);
        *mRadiusQueueEnd++ = tri;
        std::push_heap(mRadiusQueueBegin.get(), mRadiusQueueEnd, CompareSubdivTri);
      }
  }


// pop a triangle
// queue should not be empty
SubdivTri *AggrErrorsTAdaptive::PopRadiusPriorityQueue()
  {
    //assert(! gRadiusQueue.empty());
    assert(mRadiusQueueBegin.get()!=mRadiusQueueEnd);
    //SubdivTri *ret=gRadiusQueue.top();
    //gRadiusQueue.pop();
    std::pop_heap(mRadiusQueueBegin.get(), mRadiusQueueEnd, CompareSubdivTri);
    --mRadiusQueueEnd;
    return *mRadiusQueueEnd;
  }

// peek at radius of top of queue
// returns 0 if queue is empty
vec_t AggrErrorsTAdaptive::PeekTopRadiusPriorityQueue()
  {
    //if (gRadiusQueue.empty())
    while (mRadiusQueueBegin.get()!=mRadiusQueueEnd)
      {
        if ((*mRadiusQueueBegin)->IsLeaf())
          return (*mRadiusQueueBegin)->EffectiveRadius();
          //return gRadiusQueue.top()->EffectiveRadius();
        PopRadiusPriorityQueue();
      }
    return 0;
  }

inline int AggrErrorsTAdaptive::SplitsLeftInPool()
  { 
    return (mAllocatePoolSize-(mSubdivTriPoolEnd-mSubdivTriPool.get()))>>2; 
  }

inline int AggrErrorsTAdaptive::NumLeavesInPool()
  {
    return (mSubdivTriPoolEnd - mSubdivTriPool.get() + 1)>>1;
  }

void SubdivTri :: SplitHelper(int CachedHypot)
  {
    assert(IsLeaf());

    SubdivTri *left=Neighbor(Left);
    SubdivTri *right=Neighbor(Right);
    SubdivTri *child=sAdaptive->AllocateTwoSubdivTris();
      {
        int verts[3]={ Vertex(Base), Vertex(Left), HypotVertIndex() };
        child[Left].SetLeaf(verts);
        child[Left].HypotVertIndex()=gModel->BisectSafe(verts[0], verts[1]); 
        child[Left].CachedVertex(Left) = CachedVertex(Base);
        child[Left].CachedVertex(Right)= CachedVertex(Left);
        child[Left].CachedVertex(Base) = CachedHypot;
      }
      {
        int verts[3]={ Vertex(Right), Vertex(Base), HypotVertIndex() };
        child[Right].SetLeaf(verts);
        child[Right].HypotVertIndex()=gModel->BisectSafe(verts[0], verts[1]); 
        child[Right].CachedVertex(Left) = CachedVertex(Right);
        child[Right].CachedVertex(Right)= CachedVertex(Base);
        child[Right].CachedVertex(Base) = CachedHypot;
      }

    child[Left].SetNeighbor(Base, left);
    if (left)
      if (left->Neighbor(Base)==this)
        left->SetNeighbor(Base, & child[Left]);
      else
        {
          assert(left->Neighbor(Right)==this);
          left->SetNeighbor(Right, & child[Left]);
          // add child[Left] to split queue
          sAdaptive->EnqueueRadiusPriorityQueue(& child[Left]);
        }
    else
      // add child[Left] to split queue
      sAdaptive->EnqueueRadiusPriorityQueue(& child[Left]);

    child[Left].SetNeighbor(Left, & child[Right]);

    child[Right].SetNeighbor(Base, right);
    if (right)
      if (right->Neighbor(Base)==this)
        right->SetNeighbor(Base, & child[Right]);
      else
        {
          assert(right->Neighbor(Left)==this);
          right->SetNeighbor(Left, & child[Right]);
          sAdaptive->EnqueueRadiusPriorityQueue(& child[Right]);
        }
    else 
      // add child[Right] to split queue
      sAdaptive->EnqueueRadiusPriorityQueue(& child[Right]);
    child[Right].SetNeighbor(Right, & child[Left]);

    SetBranch(child);
  }

void SubdivTri :: SplitEdge(int side)
  {
    if (IsLeaf())
      {
        assert(Neighbor(side)==0);
        if (side==Base)
          Split();
        else
          {
            bool ret=Split(sAdaptive->SplitsLeftInPool());
            assert(ret);
            Child(side).Split();
          }
      }
    else
      {
        if (side==Base)
          {
            Child(Left).SplitEdge(Right);
            Child(Right).SplitEdge(Left);
          }
        else
          {
            Child(side).SplitEdge(Base);
          }
      }
  }

// adaptive subdivision ----------------

AggrErrorsTAdaptive::AggrErrorsTAdaptive(int aggr) : AggrErrorsT(aggr) 
  {
    sAdaptive=this;
    mAllocatePoolSize = mAggrNumTri<<1;
    mSubdivTriPool.reset(new SubdivTri[mAllocatePoolSize]);
    mRadiusQueueBegin.reset(new SubdivTri *[mAggrNumTri<<1]);
    InitLowDetailErrorThreshold();
  }


vec_t AggrErrorsTAdaptive::AdaptiveTreeHelper(ind lind, ind rind, ind bind, int &numtri)
  {
    // Create a single leaf triangle
    int verts[3]={lind, rind, bind};
    mSubdivTriPool->SetLeaf(verts);
    mSubdivTriPool->SetNeighbor(SubdivTri::Left, 0);
    mSubdivTriPool->SetNeighbor(SubdivTri::Right, 0);
    mSubdivTriPool->SetNeighbor(SubdivTri::Base, 0);
    mSubdivTriPool->HypotVertIndex()=gModel->BisectSafe(lind,rind);

    if (gIsCaching)
      {
        vec3_t CachedVertPos;
        gModel->GetPosition(lind, CachedVertPos);
        Index first=AllocateVertexNum(CachedVertPos);
        mSubdivTriPool->CachedVertex(SubdivTri::Left)=first;
        gModel->GetPosition(rind, CachedVertPos);
        AllocateVertex(CachedVertPos);
        mSubdivTriPool->CachedVertex(SubdivTri::Right)=first+1;
        gModel->GetPosition(bind, CachedVertPos);
        AllocateVertex(CachedVertPos);
        mSubdivTriPool->CachedVertex(SubdivTri::Base)=first+2;
      }

    // priority queue of Radii?
    InitRadiusPriorityQueue(mSubdivTriPool.get());

    mSubdivTriPoolEnd=mSubdivTriPool+1;

#ifndef CRACKS
    // Split until boundary correct
    int i;
    for (i=0; i<mAggregation; ++i)
      {
        mSubdivTriPool->SplitEdge(SubdivTri::Base);
        mSubdivTriPool->SplitEdge(SubdivTri::Left);
        mSubdivTriPool->SplitEdge(SubdivTri::Right);
      }
#endif

    int remainder = SplitsLeftInPool();
    while (remainder)
      {
        SubdivTri *tri = PopRadiusPriorityQueue();
        if (tri->IsLeaf())
          {
            if (! tri->Split(remainder))
              {
                numtri = NumLeavesInPool();
                return tri->EffectiveRadius();
              }
          }
        remainder = SplitsLeftInPool();
      }

    numtri = NumLeavesInPool();

    // radius comes from top of queue
    // if queue empty return 0;
    return PeekTopRadiusPriorityQueue();
  }

int AggrErrorsTAdaptive::ConstructAdaptiveTree(ind lind, ind rind, ind bind)
  { 
    int numtri=0;
    AdaptiveTreeHelper(lind, rind, bind, numtri);
    return numtri;
  }

vec_t AggrErrorsTAdaptive::RadiusForAdaptiveTree(ind lind, ind rind, ind bind)
  { 
    int numtri=0;
    return AdaptiveTreeHelper(lind, rind, bind, numtri);
  }

void AggrErrorsTAdaptive::RenderAdaptiveTree()
  {
    // Render the tree
    mSubdivTriPool->Render0();
  }
