// copyright 2002 Josh Levenberg, please see copying.txt
/* OpenGL code for rendering blocks of geometry.  A VertexArray cache 
   consists of an array of vertices and an array of indices -- together 
   these describe a block of geometry in a suitable format for 
   glDrawElements.
*/
#include "disp_va.h"
#include "dlod.h"
#include <assert.h>

void VertexArrayDisplay::Initialize(int maxnumvert, int maxnumind)
  {
    assert(mVertices==0);
    assert(mIndices==0);
    assert(mFirstCache==0);
    assert(mFreeCache==0);
    assert(mCurrentCache==0);
    mBlockSize=maxnumvert;
    mDefaultTriMemory=maxnumind;
    mVertices=0;
    mIndices=0;
  }

void VertexArrayDisplay::Done()
  {
    assert(mVertices==0);
    assert(mIndices==0);
    CacheType tofree=mFreeCache;
    while (tofree)
      {
        CacheType next=tofree->mNext;
        assert(tofree->mVertices==0);
        assert(tofree->mIndices==0);
        delete tofree;
        tofree=next;
      }
    mFreeCache=0;
    tofree=mFirstCache;
    while (tofree)
      {
        CacheType next=tofree->mNext;
        delete[] tofree->mVertices;
        delete[] tofree->mIndices;
        delete tofree;
        tofree=next;
      }
    mFirstCache=0;
  }

bool VertexArrayDisplay::BeginCache(Coord *&verts, Index *&inds)
  {
    assert(mVertices==0);
    assert(mIndices==0);

    verts=mVertices=new float[mBlockSize*FloatsPerVert];
    inds=mIndices=new indextype[mDefaultTriMemory];

    CacheType ret=mFreeCache;
    if (ret)
      mFreeCache=ret->mNext;
    else
      {
        ret = new cachestruct;
      }
    ret->mNext=mFirstCache;
    ret->mPrev=0;
    if (mFirstCache)
      mFirstCache->mPrev=ret;
    mFirstCache=ret;
    ret->mVertices = mVertices;
    ret->mIndices  = mIndices;
#ifdef RANGE_ELEMENTS_VA
    ret->mNumVertices=0;
#endif
    ret->mNumIndices=0;
    mCurrentCache=ret;

    glVertexPointer(3, GL_FLOAT, FloatsPerVert*sizeof(float), mVertices);
    glTexCoordPointer(2, GL_FLOAT, FloatsPerVert*sizeof(float), mVertices);
    return true;
  }

Cache VertexArrayDisplay::FinishCache(int numvert, int numind)
  {
    assert(mVertices);
    assert(mIndices);
    assert(numvert<=mBlockSize);
    assert(numind<=mDefaultTriMemory);
    assert(IsValidCache(mCurrentCache));

#ifdef RANGE_ELEMENTS_VA
    /*
      numvert==153
      numind==768
      indexenum==5123
      mVertices && mIndices &&  
      numvert<=mBlockSize && 
      numind<=mDefaultTriMemory &&
      IsValidCache(mCurrentCache) == 1
      min of mIndices..mIndices+numind==0
      max of ditto==152
    */
    glDrawRangeElements(GL_TRIANGLES, 0, numvert,
                        numind, indexenum, mIndices);
    mCurrentCache->mNumVertices = numvert;
#else
    glDrawElements(GL_TRIANGLES, numind, indexenum, mIndices);
#endif
    mCurrentCache->mNumIndices = numind;

    mVertices=0;
    mIndices=0;

    Cache ret = reinterpret_cast<Cache>(mCurrentCache);
    mCurrentCache=reinterpret_cast<CacheType>(InvalidCache);
    return ret;
  }

void VertexArrayDisplay::RenderCache(Cache c)
  {
    CacheType cache=reinterpret_cast<CacheType>(c);
    assert(IsValidCache(cache));
    assert(!IsValidCache(mCurrentCache));

    glVertexPointer(3, GL_FLOAT, FloatsPerVert*sizeof(float), cache->mVertices);
    glTexCoordPointer(2, GL_FLOAT, FloatsPerVert*sizeof(float), cache->mVertices);

#ifdef RANGE_ELEMENTS_VA
    glDrawRangeElements(GL_TRIANGLES, 0, cache->mNumVertices,
                        cache->mNumIndices, indexenum, cache->mIndices);
#else
    glDrawElements(GL_TRIANGLES, cache->mNumIndices, indexenum, cache->mIndices);
#endif
  }

void VertexArrayDisplay::DiscardCache(Cache c)
  {
    CacheType cache=reinterpret_cast<CacheType>(c);
    assert(IsValidCache(cache));
    assert(!IsValidCache(mCurrentCache));
    if (cache->mPrev)
      {
        cache->mPrev->mNext = cache->mNext;
        if (cache->mNext)
          cache->mNext->mPrev = cache->mPrev;
        cache->mPrev = 0;
      }
    else
      {
        assert(cache==mFirstCache);
        if (cache->mNext)
          cache->mNext->mPrev = 0;
        mFirstCache=cache->mNext;
      }
    cache->mNext=mFreeCache;
    mFreeCache=cache;
    delete[] cache->mVertices;
    cache->mVertices=0;
    delete[] cache->mIndices;
    cache->mIndices=0;
  }

void VertexArrayDisplay::CheckBlockConsistency()
  {
    CacheType tofree=mFreeCache;
    while (tofree)
      {
        assert(tofree->mVertices==0);
        assert(tofree->mIndices==0);
        tofree=tofree->mNext;
      }
    tofree=mFirstCache;
    while (tofree)
      {
        assert(tofree->mVertices);
        assert(tofree->mIndices);
        tofree=tofree->mNext;
      }
  }

