// copyright 2002 Josh Levenberg, please see copying.txt

/*
  A generic interface to the rendering subsystem (abstract base class).

  Deals with blocks of geometry called "caches".  The maximum size for
  caches is defined when the display is initialized.  Caches are rendered
  when they are created (using BeginCache/FinishCache).  After the cache
  is created, it may be refered to using the cache identifier (with type 
  "Cache") returned by FinishCache.  It is presumed that rendering an 
  existing cache with RenderCache is faster than creating a new cache.  
  Use DiscardCache when the application is done with the cache.  After
  calling DiscardCache, the cache identifier is no longer valid.

  Typically there is a single instance of (some descendant of) 
  AbstractDisplay.  This instance is created with either
    SelectVertexArrayDisplay() or SelectNVDisplay()
  and stored in gDisplay.  When you are done with a display, call
  its Done() method.  It may then be either deleted or initialized
  with a new maximum cache size.

  Ex:

    SelectNVDisplay();
    gDisplay->Initialize(256, 6*256);
    Coord *verts;
    Index *inds;
    if (gDisplay->BeginCache(verts, inds))
      {
        int numv=0, numi=0;
        verts[numv++]= ...; // up to 256 times
        inds[numi++]= ...;  // up to 6*256 times
        Cache c=gDisplay->FinishCache(numv, numi);
        // has been rendered once
        ... // set new OpenGL modelview matrix
        gDisplay->RenderCache(c);
        ... // set new OpenGL modelview matrix
        gDisplay->RenderCache(c);
        // done rendering three copies of the geometry
        gDisplay->DiscardCache(c);
        // c no longer refers to a valid cache
      }
    glutSwapBuffers();
    gDisplay->Done();
*/

#ifndef ABSTRACT_DISPLAY_H
#define ABSTRACT_DISPLAY_H

typedef float Coord;
typedef unsigned short Index;
typedef int Cache;

/*
  Format for vertex is x, y, z, followed by optional padding.
  Each component is of type "Coord".  The total size is given by 
  FloatsPerVert.

  Note that texture coordinates are u=x and v=y.

  Changing this to 4 resulted no significant change in speed
  (even though data was now 16 byte aligned).  With 4, it uses
  1/3 extra video memory.  The rendering bottleneck is probably 
  somewhere else entirely.
*/
enum { FloatsPerVert = 3 };

class AbstractDisplay
  {
public:
    // AbstractDisplay::InvalidCache will never be returned by FinishCache
    enum { InvalidCache = -1 };
    static bool IsValidCache(Cache c) { return c!=InvalidCache; }

    virtual ~AbstractDisplay() { }
    virtual void Initialize(int maxnumvert, int maxnumind) = 0;
    virtual void Done() = 0;

    // Cached geometry

    // BeginCache: sets verts/inds to point to an area of memory big enough
    // for maxnumvert vertices/maxnumind indices (set in Initialize).
    // Returns true if enough memory was available to create a cache.
    virtual bool BeginCache(Coord *&verts, Index *&inds) = 0;
    // Must be called exactly once after each successful BeginCache() call.
    // numvert and numind say how many verts/indices were actually used.
    // numvert/numind must be <= maxnumvert/maxnumind
    virtual Cache FinishCache(int numvert, int numind) = 0;
    virtual void RenderCache(Cache) = 0;
    virtual void DiscardCache(Cache) = 0;
  };

extern AbstractDisplay * gDisplay;

/* These allocate a new display and set gDisplay.
   Will first delete gDisplay if nonzero.
 */
  // VertexArrayDisplay can be used on most OpenGL hardware.  This is
  // the "non-caching" case (though RenderCache ends up being a simple
  // glDrawElements() call).
void SelectVertexArrayDisplay();
  // Requires NVidia specific extensions; caches vertices (but not
  // indices) in video memory.
void SelectNVDisplay();

  // Calls SelectNVDisplay if the NVidia extensions are available,
  // otherwise calls SleectVertexArrayDisplay.
  // Will support ATIDisplays at some point...
void SelectFastDisplay();

#endif