// copyright 2002 Josh Levenberg, please see copying.txt
/* Functions for initializing the OpenGL state, including loading & binding
  the texture map, checking for OpenGL extensions, etc.
 */
#include "dlod.h"
#include "glinit.h"
#include "pngwrap.h"
#include "glext.h"
#include <auto_arr.h>
#include <assert.h>

static std::auto_array<unsigned char> sLandscapeTexture;
static int sLandscapeTextureComponents;
static int sLandscapeTextureSize;

#define GL_COMPRESSED_LUMINANCE_ARB   0x84EA 
#define GL_COMPRESSED_RGB_ARB   0x84ED

// -------------

bool gLockingSupported;
#ifdef _WIN32
PFNGLLOCKARRAYSEXTPROC glLockArraysEXT;
PFNGLUNLOCKARRAYSEXTPROC glUnlockArraysEXT;
#endif

bool gArrayRangeSupported;
bool gFenceSupported;
bool gNVExtensions;

#ifdef _WIN32
PFNWGLAllocateMemoryNVPROC       wglAllocateMemoryNV;
PFNWGLFreeMemoryNVPROC           wglFreeMemoryNV;
PFNGLVertexArrayRangeNVPROC      glVertexArrayRangeNV;
PFNGLFlushVertexArrayRangeNVPROC glFlushVertexArrayRangeNV;

PFNGLGenFencesNVPROC             glGenFencesNV;
PFNGLSetFenceNVPROC              glSetFenceNV;
PFNGLTestFenceNVPROC             glTestFenceNV;
PFNGLFinishFenceNVPROC           glFinishFenceNV;
PFNGLDeleteFencesNVPROC          glDeleteFencesNV;
PFNGLIsFenceNVPROC               glIsFenceNV;
PFNGLGetFenceivNVPROC            glGetFenceivNV;
#endif

//bool gDrawRangeElementsSupported = false;
#ifdef _WIN32
PFNGLDrawRangeElementsPROC glDrawRangeElements;
#endif

static bool sTextureCompressionSupported;

static void InitTextureMap(const unsigned char *gray, int wid, int high, int name)
  {
    // write gray to OpenGL as a texture map
    glBindTexture(GL_TEXTURE_2D, name);

    glPixelStorei (GL_UNPACK_ALIGNMENT, 1); 

    //                                              GL_LUMINANCE?
    glTexImage2D(GL_TEXTURE_2D, 0, 1, wid, high, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, gray);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); // No filtering

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
  }

void LoadTextureMap(const char *file_name, int name)
  {
    int width, height;
    unsigned char *data;
    bool ret=read_png_8gray(file_name, &width, &height, &data);
    if (ret!=OK)
      printf("Error loading %s\n", file_name);
    InitTextureMap(data, width, height, name);
    delete[] data;
  }


void GenerateTextureMap(int heightfieldSize, float *heightfield)
  {
    sLandscapeTexture.reset(new unsigned char[heightfieldSize*heightfieldSize]);
    sLandscapeTextureSize=heightfieldSize;
    sLandscapeTextureComponents=1;

#define HEIGHT(x,y) heightfield[(y)*(heightfieldSize+1)+(x)]

    vec3_t vn;
    float dx=1.f/heightfieldSize;
    for (int y=0; y<heightfieldSize; ++y)
      for (int x=0; x<heightfieldSize; ++x)
        {
          float v1 = HEIGHT(x+1,y+1)-HEIGHT(x,y);
          float v2 = HEIGHT(x+1,y)-HEIGHT(x,y+1);

          // cross product to find normal
          vn[0] = dx*v2+v1*dx;
          vn[1] = v1*dx-dx*v2;
          vn[2] = -2*dx*dx;

          // make normal have unit length
          vnormalize(vn);

          // dot product with negative of light vector (1,1,1)/sqrt(3) to find color
          float clr = (-vn[0]-vn[1]-vn[2])/1.7321;
          if (clr<0)
            clr=0;

          // Use 20% ambient, convert to 0-255
          sLandscapeTexture[y*heightfieldSize+x]=(.8*clr+.2)*255;
        }
#undef HEIGHT
  }

void LoadLandscapeTexture(const char *filename)
  {
    int width, height;
    unsigned char *ptr;
    printf("Loading texture map %s.\n", filename);
    bool good=read_png(filename, &width, &height, &sLandscapeTextureComponents, &ptr);
    if (good)
      {
        sLandscapeTexture.reset(ptr);
        if (width!=height || (width & (width-1)))
          {
            printf("%s has bad image dimensions %d x %d\n", filename, width, height);
            good=false;
          }
        sLandscapeTextureSize=width;
      }
    else
      printf("Error reading %s\n", filename);
    if (! good)
      exit(1);
  }

void BindLandscapeTexture()
  {
    // write sLandscapeTexture to OpenGL as a texture map
    glBindTexture(GL_TEXTURE_2D, 1);

#define MIPMAPS
#ifdef MIPMAPS

    if (sTextureCompressionSupported)
      {
        glTexImage2D(GL_TEXTURE_2D, 0, sLandscapeTextureComponents, 
                    sLandscapeTextureSize, sLandscapeTextureSize, 0, 
                    sLandscapeTextureComponents==1 ? GL_COMPRESSED_LUMINANCE_ARB 
                    : GL_COMPRESSED_RGB_ARB, 
                    GL_UNSIGNED_BYTE, sLandscapeTexture.get());
        std::auto_array<unsigned char> dest1(new unsigned char[sLandscapeTextureSize*sLandscapeTextureSize>>2]);
        std::auto_array<unsigned char> dest2(new unsigned char[sLandscapeTextureSize*sLandscapeTextureSize>>4]);
        unsigned char *cur1=sLandscapeTexture.get();
        unsigned char *cur2=dest1.get();
        for (int i=1, s=sLandscapeTextureSize>>1; s; ++i, s>>=1)
          {
            gluScaleImage(sLandscapeTextureComponents==1 ? GL_LUMINANCE : GL_RGB,
                          s<<1, s<<1, GL_UNSIGNED_BYTE,
                          cur1, s, s, GL_UNSIGNED_BYTE, cur2);
            glTexImage2D(GL_TEXTURE_2D, i, sLandscapeTextureComponents, 
                        s, s, 0, 
                        sLandscapeTextureComponents==1 ? GL_COMPRESSED_LUMINANCE_ARB 
                        : GL_COMPRESSED_RGB_ARB, 
                        GL_UNSIGNED_BYTE, cur2);
            cur1=cur2;
            if (i&1)
              cur2=dest2.get();
            else
              cur2=dest1.get();
          }
      }
    else if (sLandscapeTextureComponents==1)
      {
        int ret=gluBuild2DMipmaps(GL_TEXTURE_2D, GL_LUMINANCE8,
                                  sLandscapeTextureSize, sLandscapeTextureSize,
                                  GL_LUMINANCE, GL_UNSIGNED_BYTE, sLandscapeTexture.get());
        if (ret)
          {
            printf("glu Error building Mipmaps\n");
            exit(1);
          }
      }
    else
      {
        assert(sLandscapeTextureComponents==3);

        int ret=gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
                                  sLandscapeTextureSize, sLandscapeTextureSize,
                                  GL_RGB, GL_UNSIGNED_BYTE, sLandscapeTexture.get());
        if (ret)
          {
            printf("glu Error building Mipmaps\n");
            exit(1);
          }
      }

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);

#else
    if (sTextureCompressionSupported)
      glTexImage2D(GL_TEXTURE_2D, 0, sLandscapeTextureComponents, 
                  sLandscapeTextureSize, sLandscapeTextureSize, 0, 
                  sLandscapeTextureComponents==1 ? GL_COMPRESSED_LUMINANCE_ARB 
                  : GL_COMPRESSED_RGB_ARB, 
                  GL_UNSIGNED_BYTE, sLandscapeTexture.get());
    else
      glTexImage2D(GL_TEXTURE_2D, 0, sLandscapeTextureComponents, 
                  sLandscapeTextureSize, sLandscapeTextureSize, 0, 
                  sLandscapeTextureComponents==1 ? GL_LUMINANCE : GL_RGB, 
                  GL_UNSIGNED_BYTE, sLandscapeTexture.get());
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);         // Linear Filtering
#endif

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);           // Texture wrapping
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  }


void APIENTRY EmulateDrawRangeElements(GLenum mode, GLuint start, 
                                       GLuint end, GLsizei count, 
                                       GLenum type, const GLvoid *indices)
  {
    glDrawElements(mode, count, type, indices);
  }

// From "All About OpenGL Extensions"
// http://www.opengl.org/developers/code/features/OGLextensions/OGLextensions.html
static bool isExtensionSupported(const char *extension)
  {
    const GLubyte *extensions = NULL;
    const GLubyte *start;
    GLubyte *where, *terminator;

    /* Extension names should not have spaces. */
    where = (GLubyte *) strchr(extension, ' ');
    if (where || *extension == '\0')
      return false;
    extensions = glGetString(GL_EXTENSIONS);
    /* It takes a bit of care to be fool-proof about parsing the
       OpenGL extensions string. Don't be fooled by sub-strings,
       etc. */
    start = extensions;
    for (;;)
      {
        where = (GLubyte *) strstr((const char *) start, extension);
        if (!where)
          break;
        terminator = where + strlen(extension);
        if (where == start || *(where - 1) == ' ')
          if (*terminator == ' ' || *terminator == '\0')
            return true;
        start = terminator;
      }
    return false;
  }

void InitGL()
  {
    glDisable(GL_POLYGON_OFFSET_FILL);

    glEnable(GL_TEXTURE_2D);                                                // Enable Texture Mapping
    glClearColor(1.0f, 1.0f, 1.0f, 0.5f);                                   // Black Background
    glClearDepth(1.0f);                                                     // Depth Buffer Setup
    glEnable(GL_DEPTH_TEST);                                                // Enables Depth Testing
    glDepthFunc(GL_LEQUAL);                                                 // The Type Of Depth Testing To Do
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);                      // Really Nice Perspective Calculations

#ifdef DEMO_MOVIE
    LoadTextureMap(FindFile("title.png"), 2);
#endif
#ifdef AUTOPILOT
    LoadTextureMap(FindFile("text.png"), 2);
#endif

    glDisable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);
    glShadeModel(GL_FLAT); 
    glDisable(GL_POLYGON_SMOOTH);
    glEnable(GL_CULL_FACE);   // try turning these off!
    glCullFace(GL_BACK);      // since most are front facing...
    glDisable(GL_POLYGON_STIPPLE);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE) ; 
    glDisable(GL_FOG);


    // name string: GL_EXT_compiled_vertex_array
    gLockingSupported=isExtensionSupported("GL_EXT_compiled_vertex_array");
#ifdef _WIN32
    if (gLockingSupported)
      {
        glLockArraysEXT = (PFNGLLOCKARRAYSEXTPROC) wglGetProcAddress("glLockArraysEXT");
        glUnlockArraysEXT = (PFNGLUNLOCKARRAYSEXTPROC) wglGetProcAddress("glUnlockArraysEXT");
      }
#endif

    if (!gLockingSupported)
      fprintf(gOutFile, "Locking not supported\n");

    gArrayRangeSupported=isExtensionSupported("GL_NV_vertex_array_range");
    if (gArrayRangeSupported)
      fprintf(gOutFile, "Array range supported\n");

    gFenceSupported=isExtensionSupported("GL_NV_fence");
    if (gFenceSupported)
      fprintf(gOutFile, "Fences supported\n");

    sTextureCompressionSupported=isExtensionSupported("ARB_texture_compression");

#ifdef _WIN32
    if (gArrayRangeSupported)
      {
        wglAllocateMemoryNV       = (PFNWGLAllocateMemoryNVPROC) 
                        wglGetProcAddress("wglAllocateMemoryNV");
        wglFreeMemoryNV           = (PFNWGLFreeMemoryNVPROC) 
                        wglGetProcAddress("wglFreeMemoryNV");
        glVertexArrayRangeNV      = (PFNGLVertexArrayRangeNVPROC) 
                        wglGetProcAddress("glVertexArrayRangeNV");
        glFlushVertexArrayRangeNV = (PFNGLFlushVertexArrayRangeNVPROC) 
                        wglGetProcAddress("glFlushVertexArrayRangeNV");
      }

    if (gFenceSupported)
      {
        glGenFencesNV     = (PFNGLGenFencesNVPROC) 
                        wglGetProcAddress("glGenFencesNV");
        glSetFenceNV      = (PFNGLSetFenceNVPROC) 
                        wglGetProcAddress("glSetFenceNV");
        glTestFenceNV     = (PFNGLTestFenceNVPROC) 
                        wglGetProcAddress("glTestFenceNV");
        glFinishFenceNV   = (PFNGLFinishFenceNVPROC) 
                        wglGetProcAddress("glFinishFenceNV");
        glDeleteFencesNV  = (PFNGLDeleteFencesNVPROC) 
                        wglGetProcAddress("glDeleteFencesNV");
        glIsFenceNV       = (PFNGLIsFenceNVPROC) 
                        wglGetProcAddress("glIsFenceNV");
        glGetFenceivNV    = (PFNGLGetFenceivNVPROC) 
                        wglGetProcAddress("glGetFenceivNV");
      }

    gNVExtensions = gArrayRangeSupported && gFenceSupported;

    glDrawRangeElements = (PFNGLDrawRangeElementsPROC) 
                      wglGetProcAddress("glDrawRangeElements");
    if (glDrawRangeElements==0)
      glDrawRangeElements=EmulateDrawRangeElements;
    //gDrawRangeElementsSupported = (glDrawRangeElements!=0);
#endif
  }

