// copyright 2002 Josh Levenberg, please see copying.txt

/*
  Defines 3 classes: ModelT, ErrorsT, and AggrErrorsT

  ModelT stores a height field.

  ErrorT stores the error for using a diamond from a BTT instead of the full
  heightfield resolution.  Errors are indexed by the midpoint of the edge
  shared by the two triangles.  Errors are represented by the distance the
  camera has to be from the midpoint for the approximation to be within the
  error tolerance.  ErrorTs are created by ModelTs given an error ratio
  (the camera distance to world space error ratio).

  AggrErrorT stores the error for using an aggregate triangle.  It also
  has methods for creating and rendering aggregate triangles.  AggrErrorTs
  are created from ErrorTs given an aggregation level and whether we
  are to use uniformly or adaptively subdivided aggregate triangles.
*/
#ifndef MODEL_H
#define MODEL_H

#include "dlod.h"
#include "jstring.h"
#include <assert.h>
#include <auto_arr.h>
//#include <memory>
//#include <vector>
#include <boost/utility.hpp>

class ModelT;
class ErrorsT;
class AggrErrorsT;

typedef ModelT Model;
typedef ErrorsT Errors;
typedef AggrErrorsT AggrErrors;

// these should maybe be auto_ptrs
extern Model *gModel;
extern Errors *gErrors;
extern AggrErrors *gAggrErrors;

void InitializeModel(std::auto_array<unsigned short> hf, vec_t maxH, int shift);
void InitializeErrors(vec_t error_ratio);
void InitializeAggregation(int aggr, bool adapt);


class TCommon : boost::noncopyable
  {
protected:
    typedef ModelT Model;
    typedef ErrorsT Errors;
    typedef int ind;  // an index indicating a vertex position
    typedef int vec;  // delta between to ind's
    enum { InvalidInd = -1 };
  };

class ModelT : TCommon
  {
protected:
    friend class ErrorsT;
    friend class AggrErrorsT;

    std::auto_array<unsigned short> mH;
    int mShift;
    int mSize;
    int mMask;
    vec_t mMaxH;
    vec_t mHScale;
    vec_t mXYScale;

    inline int GetVertexIndex(int x, int y)
      {
        assert(x>=0);
        assert(x<=mSize);
        assert(y>=0);
        assert(y<=mSize);
        return (y<<(mShift+1))+x;
      }

    inline short Xof(ind i)
      {
        return i & mMask;
      }
  
    inline short Yof(ind i)
      {
        return i >> (mShift+1);
      }

#ifndef NDEBUG
    inline bool GoodIndex(ind i)
      {
        return Xof(i)>=0 && Xof(i)<=mSize && Yof(i)>=0 && Yof(i)<=mSize && 
              i==GetVertexIndex(Xof(i), Yof(i));
      }

    inline bool GoodVec(vec v)
      {
        return v>=-GetVertexIndex(mSize,mSize) && 
               v<= GetVertexIndex(mSize,mSize);
      }
#endif

    inline void GetPosition(int x, int y, vec3_t loc)
      {
        assert(x>=0);
        assert(x<=mSize);
        assert(y>=0);
        assert(y<=mSize);
        loc[0]=x*mXYScale;
        loc[1]=y*mXYScale;
        loc[2]=mH[y*(mSize+1)+x]*mHScale;
      }

    inline int ArrayIndex(ind v)
      { assert(GoodIndex(v));
        int x=Xof(v);
        int y=Yof(v);
        return y*(mSize+1)+x;
      }

public:
    enum { InvalidInd= TCommon::InvalidInd };
    enum { Left, Right, Base };
    ModelT(std::auto_array<unsigned short> hf, vec_t maxH, int shift)
      : mH(hf), mShift(shift), mMaxH(maxH), mSize(1<<shift), 
        mMask((1<<(shift+1))-1), mHScale(maxH/((1<<16)-1)), mXYScale(1./(1<<shift))
      { }

    int GetNumBaseTriangles()
      {
        return 2;
      }

    // returns -1 if none
    int GetAdjacentBaseTriangle(int tri, int dir)
      {
        if (dir==Base)
          return tri ^ 1;
        else
          return -1;
      }

    int GetMaxSubdivision()
      {
        return mShift*2;
      }

    // assumes can bisect with (v1+v2)>>1
    // but can only bisect GetMaxSubdivision() times
    ind GetBaseTriangleVert(int tri, int vert)
      {
        static int v[6][2] =
          { {0, 0}, {1, 1}, {0, 1},
            {1, 1}, {0, 0}, {1, 0} };
        int i=tri*3+vert;
        return GetVertexIndex(v[i][0] << mShift, v[i][1] << mShift);
      }

    // index must point to num elements
    // output must point to 3*num elements
    // index[0..num-1] must be from the same base triangle
    //Not used yet: void IndexToLocation(int num, const ind *index, vec_t *output);
    inline void GetPosition(ind vert, vec3_t loc)
      {
        assert(GoodIndex(vert));
        int x=Xof(vert);
        int y=Yof(vert);
        loc[0]=x*mXYScale;
        loc[1]=y*mXYScale;
        loc[2]=mH[y*(mSize+1)+x]*mHScale;
      }

    vec_t LookUpHeight(vec_t x, vec_t y);

    std::auto_ptr<Errors> CalculateErrors(vec_t error_ratio) const;

    // something for texture coordinates (I currently assume u=x, v=y)

    // return some sort of bounding volume
    // virtual ??? TriangleBound(int base_tri, unsigned verts[3]) = 0;
    // better yet, a bound given three verts, sets a vector<vec3_t>
    void BoundingVolume(ind left, ind right, ind base, 
                        std::vector<vec3s_t> &ret);

    void BoundingVolume(int basetri, std::vector<vec3s_t> &ret)
      {
        BoundingVolume(GetBaseTriangleVert(basetri, Left), 
                       GetBaseTriangleVert(basetri, Right), 
                       GetBaseTriangleVert(basetri, Base), ret);
      }


    // FUTURE-proof:
    bool Cull(int base_tri, float camera_position[3])
      {
        return false;
      }
    // returns true if base triangle invisible/occluded from camera_position

    // return InvalidInd if too close
    inline ind BisectSafe(ind i1, ind i2)
      {
        assert(GoodIndex(i1));
        assert(GoodIndex(i2));
        int test=abs(i1-i2);
        if (test==mMask || test==mMask+2)
          return InvalidInd;
        else
          return (i1+i2)>>1;
      }

    static inline ind BisectUnsafe(ind i1, ind i2)
      {
        assert(gModel->GoodIndex(i1));
        assert(gModel->GoodIndex(i2));
        // check for accidental truncation
        assert(gModel->BisectSafe(i1, i2)!=-1);
        return (i1+i2)>>1;
      }

    static inline ind Add(ind i, vec v)
      {
        assert(gModel->GoodIndex(i));
        assert(gModel->GoodVec(v));
        return i+v;
      }

    static inline ind Sub(ind i, vec v)
      {
        assert(gModel->GoodIndex(i));
        assert(gModel->GoodVec(v));
        return i-v;
      }

    static inline vec Diff(ind i1, ind i2)
      {
        assert(gModel->GoodIndex(i1));
        assert(gModel->GoodIndex(i2));
        return i1-i2;
      }

    static inline vec Halve(vec v)
      { // assumes not truncating
        assert(gModel->GoodVec(v));
        return v>>1;
      }
  };

class ErrorsT : TCommon
  {
protected:
    std::auto_array<vec_t> mErrorThreshold;
    vec_t mErrorRatio;
    int mSize; // copy from ModelT

    vec_t ErrorThresholdHelper(vec_t xn, vec_t yn, vec_t offset, vec3_t refpt,
                           int xb, int yb, int xl, int yl, int xr, int yr);
    void ComputeErrorThreshold(int xb, int yb, int xl, int yl, int xr, int yr);
    void InitErrorThreshold();

    friend class ModelT;
    ErrorsT(vec_t error_ratio) : mErrorRatio(error_ratio), mSize(gModel->mSize),
        mErrorThreshold(new vec_t[(gModel->mSize+1)*(gModel->mSize+1)])
      { InitErrorThreshold(); }
public:
    vec_t GetErrorRatio()
      { return mErrorRatio; }
    vec_t Get(ind vert)
      {
        assert(gModel->GoodIndex(vert));
        return mErrorThreshold[gModel->ArrayIndex(vert)];
      }
    std::auto_ptr<AggrErrors> ComputeAggregation(int agglevel, bool adaptive) const;
  };

class AggrErrorsT : TCommon
  {
protected:
    int mAggregation;
    std::auto_array<vec_t> mLowDetailErrorThreshold;
    /* [gLowDetailTessellation+1][gLowDetailTessellation+1] */

    int mAggrNumTri;
    int mLowDetailTessellation;
    int mLowDetailTessellationShift;
    int mYShift;

    void InitLowDetailErrorThreshold();
    void ComputeLowDetailErrorThreshold(int xb, int yb, int xl, int yl, int xr, int yr);

    // returns error threshold radius
    vec_t RadiusForUniformTree(ind, ind, ind);
    virtual vec_t RadiusForTree(ind, ind, ind)=0;

    // puts tree in mSubdivTriPool
    // returns number of triangles
    int ConstructUniformTree(ind, ind, ind);
    virtual int ConstructTree(ind, ind, ind)=0; // one of the above

    // renders tree in mSubdivTriPool
    void RenderUniformTree();
    virtual void RenderTree()=0; // one of the above

    inline vec_t & Ref(int y, int x)
      {
        y >>= mAggregation;
        x >>= mAggregation;
        return mLowDetailErrorThreshold[y*(mLowDetailTessellation+1) + x];
      }

    AggrErrorsT(int aggr);

public:
    virtual ~AggrErrorsT() { }
    int GetAggregation()
      { return mAggregation; }
    virtual bool GetAdaptive() = 0;
    vec_t Get(ind vert);

    // returns number of triangles
    int RenderUniform(ind l, ind r, ind b)                  
      {
        int ret=ConstructUniformTree(GetHD(l), GetHD(r), GetHD(b));
        RenderUniformTree();
        return ret;
      }

    // returns number of triangles
    int Render(int l, int r, int b)
      {
        int ret=ConstructTree(GetHD(l), GetHD(r), GetHD(b));
        RenderTree();
        return ret;
      }


#ifndef NDEBUG
    bool GoodIndex(ind i);
#endif

    // BOGUS: remove need for BisectSafe by checking
    // outside function for error case

    // return -1 if too close
    ind BisectSafe(ind v1, ind v2)
      {
        assert(GoodIndex(v1));
        assert(GoodIndex(v2));
        int test=abs(v1-v2);
        int xdelta=1<<mAggregation;
        int cmpto=gModel->GetVertexIndex(xdelta,xdelta);
        if (test==cmpto || test==cmpto-(xdelta<<1))
          return -1;
        else
          return (v1+v2)>>1;
      }

    /* Not used yet:
    ind BisectUnsafe(ind v1, ind v2)
      {
        assert(BisectSafe(v1, v2)==((v1+v2)>>1));
        return (v1+v2)>>1;
      }
    */

    inline ind GetHD(int ld)
      {
        assert(GoodIndex(ld));
        return ld;
      }

    inline void GetPosition(int vert, vec3_t loc)
      {
        assert(GoodIndex(vert));
        gModel->GetPosition(vert, loc);
      }

    inline vec_t GetRadius(ind vert)
      {
        assert(GoodIndex(vert));
        int x = (vert >> mAggregation) & ((mLowDetailTessellation<<1)-1);
        int y = (vert >> mYShift);
        
        return mLowDetailErrorThreshold[y*(mLowDetailTessellation+1)+x];
      }

    inline int GetVertexIndex(int x, int y)
      {
        return gModel->GetVertexIndex(x, y);
      }
    ind GetBaseTriangleVert(int tri, int vert)
      { return gModel->GetBaseTriangleVert(tri, vert); }

  };

#endif